/*****************************************************************************
 * $Id: diffarch.c,v 2.20 1996/06/06 20:03:01 ak Exp $
 *****************************************************************************
 * $Log: diffarch.c,v $
 * Revision 2.20  1996/06/06 20:03:01  ak
 * Separate enablers for compression engines.
 * Better compression error checks and messages.
 * ZLIB disabled by default - unstable.
 *
 * Revision 2.19  1996/06/06 15:33:31  ak
 * ZLIB based compression engine.
 *
 * Revision 2.18  1996/06/03 01:56:13  ak
 * Support verify mode.
 * Support append/update to device, if the device supports.
 * Still no delete though.
 *
 * Revision 2.17  1996/06/02 20:05:17  ak
 * Release 2.49.
 *
 * Revision 2.16  1995/04/25 23:27:00  ak
 * Compare compressed files.
 *
 * Revision 2.15  1995/03/27  01:32:52  ak
 * Added individual file compression (--fzip). Create/extract only, no diff, no update.
 *
 * Revision 2.14  1994/11/19  00:16:17  ak
 * Disabled DUMPDIR compares, the tags are almost always different.
 * Should compare contents individually, not bcmp().
 *
 * Revision 2.13  1994/11/08  20:12:19  ak
 * Optional case-sensitive compare.
 * Moved include of port.h to tar.h.
 *
 * Revision 2.12  1994/10/29  20:58:07  ak
 * Bugfix multivolume QFA.
 *
 * Revision 2.11  1994/09/07  18:45:10  edvkai
 * AIX
 *
 * Revision 2.10  1994/07/05 18:44:55  edvkai
 * 2.35: Posix support for long filenames and magic name.
 *
 * Lots of changes in filename handling, I hope I found all places. Added
 * decode_filename() with result in var 'filename' for this. header.name
 * should no longer be used. The diffarch kludge FILENAME is no longer
 * necessary.
 *
 * Treatment of links is not up-to-date, I fear - what's defined by Posix
 * about long link names?
 *
 * Due to NONAMES, prior versions created "oldarch" style archives without
 * magic name. Now magic is "ustar\0""00" for Posix and "ustar  \0" else.
 *
 * Unused header space is used to support multi-volume archives in Posix -
 * THIS IS SPECIFIC TO GTAK.
 *
 * Note that Posix conflicts with atime/ctime.
 *
 * Revision 2.9  1994/07/05 14:56:17  edvkai
 * Access Control Lists.
 *
 * Added --substitute, allows redirection of UNC names. No support for links yet.
 *
 * Streamlined absolute path treatment. No skipcrud any more.
 *
 * Bugfix: If a directory existed, EAs were not attached.
 *
 * Revision 2.8  1994/05/31 23:23:44  ak
 * Reversed meaning of --old-ea to stay compatible to gtak 2.12. The
 * 32-bit API should have a longer life anyway (PowerPC?) and is faster.
 *
 * Avoid _ead_read like the plague (sorry E.M.). Crashed in diffarch if
 * the EMX libs where being used but worked when I compiled the patched
 * ead* sources into TAR. Now using ea_load in diffarch too.
 *
 * Removed EA_MODE conditionals.
 *
 * Revision 2.7  1994/05/31 21:33:19  ak
 * Still there seems to be a memory leak somewhere in EMX _ead_read.
 * Didn't find it with dbmalloc though. Added EA_MODE (former EMX_EA) of
 * 2, using my own ea_* functions to load the attributes and the EMX
 * functions to manipulate them. The leak seems to be gone now. Added
 * --old-ea to allow creating the gtak 2.12 EA records.
 *
 * Revision 2.6  1994/02/27 21:51:25  ak
 * Some fixes and additions in EA management.
 * Use same mangled name both for file and for EA entry.
 * Fixed serious bug in mangled name rename entry (bad st_size).
 *
 * Revision 2.5  1994/02/18 19:44:39  ak
 * Add check for added or missing EA names.
 *
 * Revision 2.4  1994/02/17 16:02:17  ak
 * Use EMX _ead_* extended attribute functions. Create new EAs (ifdef EMX_EA),
 * accept both old and new when reading. As the new representation reflects
 * the OS/2 1.x binary EA representation, GTAK 1.x EAs are accepted now.
 *
 * Revision 2.3  1993/11/25 20:23:38  edvkai
 * Major QFA changes.
 *
 * QFA now supports both block id and relative addressing. Relative
 * addressing is more efficient with short distances. 4mm and 8mm devices
 * do well with relative addressing only. For the sake of QIC tapes, TAR
 * optimizes QFA using block ids for large distances.
 *
 * The format of the tape-directory changed. No longer compatible to prior
 * versions. Maybe I'll add a utility for conversion.
 *
 * Rmt interface changed. Now using block numbers instead of byte offsets
 * for seek. Previously the lseek limitation to 2GB limited QFA archive
 * capacity. Changed rmtlseek to rmtseek to indicate that it is different.
 *
 * QFA should be able to handle multi-volume archives now.
 *
 * If a volume name is specified on the command line, QFA skips archives
 * having a different volume name.
 *
 * Code for header pretty-printing cleaned. Doesn't depend on hstat any
 * longer and requires less globals. No explanation text behind names in
 * tape directory.
 *
 * A few bugfixes.
 *
 * Revision 2.2  1993/08/08  19:27:42  ak
 * *** empty log message ***
 *
 * Revision 2.1  1993/08/08  19:07:26  ak
 * Merge of network TAR with 2.12.
 *
 * Revision 1.7  1993/01/14  21:50:37  ak
 * EMX 0.8f fixes and enhancements by Kai Uwe Rommel.
 *
 * Revision 1.6  1993/01/12  18:24:55  ak
 * setnbuf for TTY output only.
 *
 * Revision 1.5  1993/01/10  11:54:43  ak
 * diff_init sets output stream to non-buffered.
 *
 * Revision 1.1.1.2  1993/08/08  17:49:32  ak
 * - Network tape access.
 * - Dynamically loaded modules for disk/scsi/network interfaces.
 *
 * Revision 1.4  1992/12/19  22:22:24  ak
 * Eberhard Mattes: Trapped when ea_load returned NULL.
 *
 * Revision 1.3  1992/09/12  15:57:23  ak
 * - Usenet patches for GNU TAR 1.10
 * - Bugfixes and patches of Kai Uwe Rommel:
 *         filename conversion for FAT
 *         EMX 0.8e
 *         -0..1 alias for a: b:
 *         -2..7 alias for +++TAPE$x
 *
 * Revision 1.2  1992/09/02  20:07:55  ak
 * Version AK200
 * - Tape access
 * - Quick file access
 * - OS/2 extended attributes
 * - Some OS/2 fixes
 * - Some fixes of Kai Uwe Rommel
 *
 * Revision 1.1.1.1  1992/09/02  19:21:14  ak
 * Original GNU Tar 1.10 with some filenames changed for FAT compatibility.
 *
 * Revision 1.1  1992/09/02  19:21:12  ak
 * Initial revision
 *
 *****************************************************************************/

static char *rcsid = "$Id: diffarch.c,v 2.20 1996/06/06 20:03:01 ak Exp $";

/*
 * Modified by Andreas Kaiser July 92.
 * See CHANGES.AK for info.
 */

/* Modified so that the verify option works even if some filenames got
   mangled because of excessive length.  Note that this also required
   adding an inverse_find_mangled operation to mangle.c.  While I was
   at it I also fixed the handling of absolute pathnames and also
   three buggy error messages in the handling of links that used the wrong
   names.  I didn't make the verification of mangled names or absolute
   pathnames work for dump directories.
     -Max Hailperin <max@nic.gac.edu> 8/1/91 */

/* Further modified so that diff won't complain about uid or gid differing
   if the numerically stored id matches, even if the name doesn't tally.
   This prevents error messages verifying an archive with files owned by
   non-existant users or groups.
     -Max Hailperin <max@nic.gac.edu> 8/2/91 */

/* Fixed two bugs in the changes of 8/1/91.
    1) For some reason linkabspath was one byte two small, even though
       abspath was right.
    2) The filename and linkname variables were initialized to point into
       a header record in the buffer which might go away in the course of
       diffing the file contents, leading to garbage filenames in data
       differs error message.  The saverec mechamism only protects head,
       not other pointers into that record.
   -Max Hailperin <max@nic.gac.edu> 8/11/91 */

/* Diff files from a tar archive.
   Copyright (C) 1988 Free Software Foundation

This file is part of GNU Tar.

GNU Tar is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU Tar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Tar; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * Diff files from a tar archive.
 *
 * Written 30 April 1987 by John Gilmore, ihnp4!hoptoad!gnu.
 *
 * @(#) diffarch.c 1.10 87/11/11 - gnu
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef V7
#include <fcntl.h>
#endif

#ifdef BSD42
#include <sys/file.h>
#endif

#ifndef MSDOS
#include <sys/ioctl.h>
#if !defined(USG) || defined(HAVE_MTIO)
#include <sys/mtio.h>
#endif
#endif

#ifdef USG
#include <fcntl.h>
#endif

/* Some systems don't have these #define's -- we fake it here. */
#ifndef O_RDONLY
#define	O_RDONLY	0
#endif
#ifndef	O_NDELAY
#define	O_NDELAY	0
#endif

/*
 * Most people don't have a #define for this.
 */
#ifndef	O_BINARY
#define	O_BINARY	0
#endif

#ifndef S_IFLNK
#define lstat stat
#endif

extern int errno;			/* From libc.a */

#include "tar.h"
#include "rmt.h"

extern union record *head;		/* Points to current tape header */
extern struct stat hstat;		/* Stat struct corresponding */
extern int head_standard;		/* Tape header is in ANSI format */

extern void print_header();
extern void skip_file();
extern void skip_extended_headers();

extern FILE *msg_file;

int now_verifying = 0;		/* Are we verifying at the moment? */

int	diff_fd;		/* Descriptor of file we're diffing */

char	*diff_buf = 0;		/* Pointer to area for reading
					   file contents into */

char	*diff_dir;		/* Directory contents for LF_DUMPDIR */

#ifdef OS2
char *eaptr;			/* EA pointer */
#endif

int different = 0;

#define LINKNAME (linkname?linkname:head->header.linkname)

static char *linkname;		/* NULL means as in header */
extern char *inverse_find_mangled();
static char linkabspath[FILENAME_MAX+2] = "/";

/*struct sp_array *sparsearray;
int 		sp_ar_size = 10;*/
/*
 * Initialize for a diff operation
 */
diff_init()
{

	/*NOSTRICT*/
	diff_buf = (char *) valloc((unsigned)blocksize);
	if (!diff_buf) {
		msg("could not allocate memory for diff buffer of %d bytes",
			blocksize);
		exit(EX_ARGSBAD);
	}
	if (isatty(fileno(msg_file)))
		setvbuf(msg_file, NULL, _IONBF, 0);
}

#ifdef OS2

void
cmp_ea()
{
	int i, i2, count;
	_ead ead2 = _ead_create();
	pEABuf eabuf2 = ea_load(filename);
	if (eabuf2 == NULL && access(filename, 0) < 0) {
		/* if the file doesn't exist, the file entry
		   shows a message anyway, so forget EAs */
		;
	} else {
		_ead_use_fea2list(ead2, ptrFEAList(eabuf2));
		count = _ead_count(ead);
		for (i = 1; i <= count; ++i) {
			const char *name = _ead_get_name(ead, i);
			i2 = _ead_find(ead2, name);
			if (i2 >= 1) {
				int l = _ead_value_size(ead, i);
				int l2 = _ead_value_size(ead2, i2);
				const void *v = _ead_get_value(ead, i);
				const void *v2 = _ead_get_value(ead2, i2);
				if (l != l2 || memcmp(v, v2, l) != 0)
					fprintf(msg_file, "%s: EA '%s' differs\n",
						filename, name);
			} else
				fprintf(msg_file, "%s: EA '%s' not on disk\n",
					filename, name);
		}
		count = _ead_count(ead2);
		for (i2 = 1; i2 <= count; ++i2) {
			const char *name = _ead_get_name(ead2, i2);
			i = _ead_find(ead, name);
			if (i < 1)
				fprintf(msg_file, "%s: EA '%s' not in archive\n",
					filename, name);
		}
	}
	ea_free(eabuf2);
	_ead_destroy(ead2);
}

#endif

/*
 * Diff a file against the archive.
 */
void
diff_archive()
{
	register char *data;
	int check, namelen;
	int err, written;
	long offset, size;
	struct stat filestat;
	int compare_chunk();
	int compare_dir();
#ifdef OS2
	int compare_eattr();
	void *pfea2list;
#endif
	int no_op();
#ifndef MSDOS
	dev_t	dev;
	ino_t	ino;
#endif
	char *get_dir_contents();
	long from_oct();
#ifndef __EMX__
	long lseek();
#endif
        char *new_name;
        long numerical_uid, numerical_gid;

	errno = EPIPE;			/* FIXME, remove perrors */

	saverec(&head);			/* Make sure it sticks around */
	userec(head);			/* And go past it in the archive */
	decode_header(head, &hstat, &head_standard, 1);	/* Snarf fields */
	numerical_uid = from_oct(8,  head->header.uid);
	numerical_gid = from_oct(8,  head->header.gid);

        linkname = NULL;

	/* Print the record from 'head' and 'hstat' */
	if (f_verbose) {
		if(now_verifying)
			fprintf(msg_file,"Verify ");
		print_header(head);
	}

	switch (head->header.linkflag) {

	default:
		msg("Unknown file type '%c' for %s, diffed as normal file",
			head->header.linkflag, filename);

		/* FALL THRU */

	case LF_OLDNORMAL:
	case LF_NORMAL:
	case LF_SPARSE:
	case LF_CONTIG:
		/*
		 * Appears to be a file.
		 * See if it's really a directory.
		 */
		namelen = strlen(filename)-1;
		if (filename[namelen] == '/')
			goto really_dir;

		
		if(do_stat(&filestat)) {
			if (head->header.isextended)
				skip_extended_headers();
			skip_file((long)hstat.st_size);
			different++;
			goto quit;
		}

		if ((filestat.st_mode & S_IFMT) != S_IFREG) {
			fprintf(msg_file, "%s: not a regular file\n",
				filename);
			skip_file((long)hstat.st_size);
			different++;
			goto quit;
		}

		filestat.st_mode &= ~S_IFMT;
		if (filestat.st_mode != hstat.st_mode)
			sigh("mode");
#ifndef OS2
		if (filestat.st_uid  != hstat.st_uid &&
                    filestat.st_uid  != numerical_uid)
			sigh("uid");
		if (filestat.st_gid  != hstat.st_gid &&
                    filestat.st_gid  != numerical_gid)
			sigh("gid");
#endif
		if (filestat.st_mtime != hstat.st_mtime)
			sigh("mod time");
		if (head->header.linkflag != LF_SPARSE &&
				filestat.st_size != hstat.st_size) {
			sigh("size");
			skip_file((long)hstat.st_size);
			goto quit;
		}

		diff_fd = open(filename, O_NDELAY|O_RDONLY|O_BINARY);

		if (diff_fd < 0) {
			msg_perror("cannot open %s",filename);
			if (head->header.isextended)
				skip_extended_headers();
			skip_file((long)hstat.st_size);
			different++;
			goto quit;
		}

		if(f_multivol) {
			save_name=filename;
			save_totsize=hstat.st_size;
			save_sizeleft=hstat.st_size;
		}

		/*
		 * Need to treat sparse files completely differently here.
		 */
		if (head->header.linkflag == LF_SPARSE)
			diff_sparse_files(hstat.st_size);
		else 
			wantbytes((long)(hstat.st_size),compare_chunk);

		check = close(diff_fd);
		if (check < 0)
			msg_perror("Error while closing %s",filename);

		if(f_multivol)
			save_name=0;
	quit:
		break;

        case LF_NAMES:
                skip_file((long)hstat.st_size);
                break;

#ifndef MSDOS
	case LF_LINK:
		if(do_stat(&filestat))
			break;
		dev = filestat.st_dev;
		ino = filestat.st_ino;
              retry_linkstat:
		err = stat(LINKNAME, &filestat);
		if (err < 0) {
			if (errno==ENOENT) {
                          if(now_verifying &&
                             (new_name = inverse_find_mangled(LINKNAME))
                             != NULL){
                            linkname = new_name;
                            goto retry_linkstat;
                          }
                          else if(!f_absolute_paths && *LINKNAME != '/'){
                            strcpy(linkabspath+1,LINKNAME);
                            linkname = linkabspath;
                            goto retry_linkstat;
                          }
                          else{
                            if(!f_absolute_paths && linkname != NULL &&
                               *linkname == '/')
                              linkname++;
                            fprintf(msg_file, "%s: does not exist\n",LINKNAME);
                          }
			} else {
				msg_perror("cannot stat file %s",LINKNAME);
			}
			different++;
			break;
		}
		if(filestat.st_dev!=dev || filestat.st_ino!=ino) {
			fprintf(msg_file, "%s not linked to %s\n",filename,LINKNAME);
			break;
		}
		break;
#endif

#ifdef S_IFLNK
	case LF_SYMLINK:
	{
		char linkbuf[FILENAME_MAX+1];
              retry_readlink:
		check = readlink(filename, linkbuf,
				 (sizeof linkbuf)-1);
		
		if (check < 0) {
			if (errno == ENOENT) {
			  if ( now_verifying
			    && (new_name = inverse_find_mangled(filename)) != NULL){
				strcpy(filename, new_name);
				goto retry_readlink;
			  } else if (!f_absolute_paths && *filename != '/') {
			  	memmove(filename+1, filename, strlen(filename)+1);
			  	filename[0] = '/';
				goto retry_readlink;
			  } else if (!f_absolute_paths) {
			  	memmove(filename, filename+1, strlen(filename));
				fprintf(msg_file,
					"%s: no such file or directory\n",
					filename);
			  }
			} else {
				msg_perror("cannot read link %s",filename);
			}
			different++;
			break;
		}

		linkbuf[check] = '\0';	/* Null-terminate it */
              retry_symlink_check:
		if (strncmp(LINKNAME, linkbuf, check) != 0) {
                  if(now_verifying &&
                     (new_name = inverse_find_mangled(LINKNAME)) != NULL){
                    linkname = new_name;
                    goto retry_symlink_check;
                  }
                  else if(!f_absolute_paths && *LINKNAME != '/'){
                    strcpy(linkabspath+1, LINKNAME);
                    linkname = linkabspath;
                    goto retry_symlink_check;
                  }
                  else{
                    if(!f_absolute_paths && linkname != NULL &&
                       *linkname == '/')
                      linkname++;
			fprintf(msg_file, "%s: symlink differs\n",
				filename);
			different++;
		}
	}
	}
		break;
#endif

	case LF_CHR:
		hstat.st_mode |= S_IFCHR;
		goto check_node;

#ifdef S_IFBLK
	/* If local system doesn't support block devices, use default case */
	case LF_BLK:
		hstat.st_mode |= S_IFBLK;
		goto check_node;
#endif

#ifdef S_IFIFO
	/* If local system doesn't support FIFOs, use default case */
	case LF_FIFO:
		hstat.st_mode |= S_IFIFO;
		hstat.st_rdev = 0;		/* FIXME, do we need this? */
		goto check_node;
#endif

	check_node:
		/* FIXME, deal with umask */
		if(do_stat(&filestat))
			break;
		if(hstat.st_rdev != filestat.st_rdev) {
			fprintf(msg_file, "%s: device numbers changed\n", filename);
			different++;
			break;
		}
		if(hstat.st_mode != filestat.st_mode) {
			fprintf(msg_file, "%s: mode or device-type changed\n", filename);
			different++;
			break;
		}
		break;

	case LF_DUMPDIR:
		data=diff_dir=get_dir_contents(filename,0);
		if(data != NULL && *data) {
			wantbytes((long)(hstat.st_size),compare_dir);
			free(data);
		} else
			wantbytes((long)(hstat.st_size),no_op);
		/* FALL THROUGH */

	case LF_DIR:
		/* Check for trailing / */
		namelen = strlen(filename)-1;
	really_dir:
		while (namelen && filename[namelen] == '/')
			filename[namelen--] = '\0';	/* Zap / */

		if(do_stat(&filestat))
			break;
		if((filestat.st_mode&S_IFMT)!=S_IFDIR) {
			fprintf(msg_file, "%s is no longer a directory\n",filename);
			different++;
			break;
		}
		if((filestat.st_mode&~S_IFMT) != hstat.st_mode)
			sigh("mode");
		break;

	case LF_VOLHDR:
		break;

	case LF_MULTIVOL:
		namelen = strlen(filename)-1;
		if (filename[namelen] == '/')
			goto really_dir;

		if(do_stat(&filestat))
			break;

		if ((filestat.st_mode & S_IFMT) != S_IFREG) {
			fprintf(msg_file, "%s: not a regular file\n",
				filename);
			skip_file((long)hstat.st_size);
			different++;
			break;
		}

		filestat.st_mode &= ~S_IFMT;
		offset = from_oct(1+12, f_posix ? head->posix.offset : head->header.offset);
		if (filestat.st_size != hstat.st_size + offset) {
			sigh("size");
			skip_file((long)hstat.st_size);
			different++;
			break;
		}

		diff_fd = open(filename, O_NDELAY|O_RDONLY|O_BINARY);

		if (diff_fd < 0) {
                        msg_perror("cannot open file %s",filename);
			skip_file((long)hstat.st_size);
			different++;
			break;
		}
		err = lseek(diff_fd, offset, 0);
		if(err!=offset) {
			msg_perror("cannot seek to %ld in file %s",offset,filename);
			different++;
			break;
		}

		wantbytes((long)(hstat.st_size),compare_chunk);

		check = close(diff_fd);
		if (check < 0) {
			msg_perror("Error while closing %s",filename);
		}
		break;

#ifdef OS2
	case LF_EATTR:
		if (eabuf)
			ea_free(eabuf);
		ealen = hstat.st_size;
		eabuf = ea_alloc(hstat.st_size);
		eaptr = (char _far *)eabuf;

		if (eabuf == (pEABuf)0)
			msg("Not enough memory for extended attributes (%ld bytes)", ealen);

		for (size = ealen; size > 0; size -= RECORDSIZE) {
			union record *x = findrec();
			if (x == NULL) {
				msg("Unexpected EOF on archive file");
				break;
			}
			if (eabuf) {
				int n = (size < RECORDSIZE) ? size : RECORDSIZE;
				memcpy(eaptr, x->charptr, n);
				eaptr += n;
			}
			userec(x);
		}

		if (TestEAMagic(eabuf) || *(long *)eabuf == 0) {
			_ead_use_fea2list(ead, ptrFEAList(eabuf));
			cmp_ea(filename);
		} else if (pfea2list = _ead_fealist_to_fea2list(eabuf)) {
			_ead_use_fea2list(ead, pfea2list);
			cmp_ea(filename);
			free(pfea2list);
		}
		break;

	case LF_ACL:
	    {	char *buf = f_acl ? (char *)valloc(hstat.st_size) : NULL, *ptr = buf;
		for (size = hstat.st_size; size > 0; size -= RECORDSIZE) {
			union record *x = findrec();
			if (x == NULL) {
				msg("Unexpected EOF on archive file");
				break;
			}
			if (buf) {
				int n = (size < RECORDSIZE) ? size : RECORDSIZE;
				memcpy(ptr, x->charptr, n);
				ptr += n;
			}
			userec(x);
		}
		if (f_acl) {
			if (acl_load(acl, NULL, filename) == 0) {
				acl_bin2text(acl);
				if (acl->textlen != hstat.st_size) {
					sigh("ACL size");
					different++;
				} else if (memcmp(acl->text, buf, acl->textlen)) {
					sigh("ACL");
					different++;
				}
			} else {
				fprintf(msg_file, "%s: no ACL on disk\n", filename);
				different++;
			}
		}
		if (buf)
			free(buf);
	    }
#endif

#ifdef FZIP
	case LF_COMPR:
	    {	char *ptr;

		if (hstat.st_size > fzip_bufsize) {
		    if (fzip_buffer)
			free(fzip_buffer);
		    fzip_bufsize = (hstat.st_size | 0x3FFFF) + 1;
		    fzip_buffer = malloc(fzip_bufsize);
		    if (fzip_buffer == NULL) {
			msg("Not enough memory for decompression buffer,"
			    " extracting compressed imaged instead");
			fzip_bufsize = 0;
			skip_file((long)hstat.st_size);
			break;
		    }
		}

		diff_fd = open(filename, O_NDELAY|O_RDONLY|O_BINARY);
		if (diff_fd < 0) {
			msg_perror("cannot open %s",filename);
			if (head->header.isextended)
				skip_extended_headers();
			skip_file((long)hstat.st_size);
			different++;
			goto quit;
		}

		ptr = fzip_buffer;
		for (size = hstat.st_size; size > 0; size -= RECORDSIZE) {
		    union record *x = findrec();
		    if (x == NULL) {
			msg("Unexpected EOF on archive file");
			break;
		    }
		    memcpy(ptr, x->charptr, RECORDSIZE);
		    ptr += RECORDSIZE;
		    userec(x);
		}
		
		if(f_multivol) {
			save_name=filename;
			save_totsize=hstat.st_size;
			save_sizeleft=hstat.st_size;
		}

		err = unzip_compare(diff_fd, fzip_buffer, hstat.st_size);
		switch (err) {
		case -2:
		    msg("unknown compression used for file %s", filename);
		    break;
		case -1:
		    msg_perror("error while decompressing %s", filename);
		    break;
		case 0:
		    fprintf(msg_file, "%s: differs\n", filename);
		    different++;
		    break;
		}

		close(diff_fd);
		break;
	    }
#endif
	}

	/* We don't need to save it any longer. */
	saverec((union record **) 0);	/* Unsave it */
}

int
compare_chunk(bytes,buffer)
long bytes;
char *buffer;
{
	int err;

	err=read(diff_fd,diff_buf,bytes);
	if(err!=bytes) {
		if(err<0) {
			msg_perror("can't read %s",filename);
		} else {
			fprintf(msg_file,"%s: could only read %d of %d bytes\n",filename,err,bytes);
		}
		different++;
		return -1;
	}
	if(bcmp(buffer,diff_buf,bytes)) {
		fprintf(msg_file, "%s: data differs\n",filename);
		different++;
		return -1;
	}
	if (save_name)
		save_sizeleft -= bytes;
	return 0;
}

int
compare_dir(bytes,buffer)
long bytes;
char *buffer;
{
#if 0	/* AK - doesn't work since the tags are different */
	if(bcmp(buffer,diff_dir,bytes)) {
		fprintf(msg_file, "%s: data differs\n",filename);
		different++;
		return -1;
	}
#endif
	diff_dir+=bytes;
	return 0;
}

int
compare_eattr(bytes,buffer)
int bytes;
char *buffer;
{
#ifdef OS2
	int err = bcmp(buffer, eaptr, bytes);
	eaptr += bytes;
	if (err) {
		fprintf(msg_file, "%s: extended attribute data differs\n",filename);
		different++;
		return -1;
	}
#endif
	return 0;
}

/*
 * Sigh about something that differs.
 */
sigh(what)
	char *what;
{
	fprintf(msg_file, "%s: %s differs\n",
		filename, what);
}

verify_volume()
{
	int status;
#ifdef MTIOCTOP
	struct mtop t;
	int er;
#endif

	if(!diff_buf)
		diff_init();
#ifdef MTIOCTOP
	t.mt_op = MTBSF;
	t.mt_count = 1;
	if((er=rmtioctl(archive,MTIOCTOP,&t))<0) {
		if(errno!=EIO || (er=rmtioctl(archive,MTIOCTOP,&t))<0) {
#endif
			if (rmtseek(archive,0L,0)!=0) {
				/* Lseek failed.  Try a different method */
				msg_perror("Couldn't rewind archive file for verify");
				return;
			}
#ifdef MTIOCTOP
		}
	}
#endif
	ar_reading=1;
	now_verifying = 1;
	fl_read();
	for(;;) {
		status = read_header();
		if(status==0) {
			unsigned n;

			n=0;
			do {
				n++;
				status=read_header();
			} while(status==0);
			msg("VERIFY FAILURE: %d invalid header%s detected!",n,n==1?"":"s");
		}
		if(status==2 || status==EOF)
			break;
		decode_filename(head);
		diff_archive();
	}
	ar_reading=0;
	now_verifying = 0;

}

int do_stat(statp)
struct stat *statp;
{
	int err;
        char *new_name;

      retry:
	err = f_follow_links ? stat(filename, statp) : lstat(filename, statp);
	if (err < 0) {
		if (errno==ENOENT) {
		  if (now_verifying && (new_name = inverse_find_mangled(filename)) != NULL) {
			strcpy(filename, new_name);
			goto retry;
		  } else if (!f_absolute_paths && *filename != '/') {
			memmove(filename+1, filename, strlen(filename)+1);
			filename[0] = '/';
			goto retry;
		  } else if (!f_absolute_paths) {
			memmove(filename, filename+1, strlen(filename));
			fprintf(msg_file, "%s: does not exist\n",filename);
		  }
		} else
			msg_perror("can't stat file %s",filename);
/*		skip_file((long)hstat.st_size);
		different++;*/
		return 1;
	}
#ifdef MSDOS
	get_fileattr(filename, statp);
#endif
	return 0;
}

/*
 * JK
 * Diff'ing a sparse file with its counterpart on the tar file is a 
 * bit of a different story than a normal file.  First, we must know
 * what areas of the file to skip through, i.e., we need to contruct
 * a sparsearray, which will hold all the information we need.  We must
 * compare small amounts of data at a time as we find it.  
 */

diff_sparse_files(filesize)
int	filesize;

{
	int		sparse_ind = 0;
	char		*buf;
	int		buf_size = RECORDSIZE;
	union record 	*datarec;	
	int		err;
	long		numbytes;
	int		amt_read = 0;
	int		size = filesize;

	buf = (char *) malloc(buf_size * sizeof (char));
	
	fill_in_sparse_array();
	

	while (size > 0) {
		datarec = findrec();
		if (!sparsearray[sparse_ind].numbytes)
			break;

		/*
		 * 'numbytes' is nicer to write than
		 * 'sparsearray[sparse_ind].numbytes' all the time ...
		 */
		numbytes = sparsearray[sparse_ind].numbytes;
		
		lseek(diff_fd, sparsearray[sparse_ind].offset, 0);
		/*
		 * take care to not run out of room in our buffer
		 */
		while (buf_size < numbytes) {
			buf = (char *) realloc(buf, buf_size * 2 * sizeof(char));
			buf_size *= 2;
		}
		while (numbytes > RECORDSIZE) {
			if ((err = read(diff_fd, buf, RECORDSIZE)) != RECORDSIZE) {
	 			if (err < 0) 
					msg_perror("can't read %s", filename);
				else
					fprintf(msg_file, "%s: could only read %d of %d bytes\n", 
						err, numbytes);
				break;
			}
			if (bcmp(buf, datarec->charptr, RECORDSIZE)) {
				different++;
				break;
			}
			numbytes -= err;
			size -= err;
			userec(datarec);
			datarec = findrec();
		}
		if ((err = read(diff_fd, buf, numbytes)) != numbytes) {
 			if (err < 0) 
				msg_perror("can't read %s", filename);
			else
				fprintf(msg_file, "%s: could only read %d of %d bytes\n", 
						err, numbytes);
			break;
		}

		if (bcmp(buf, datarec->charptr, numbytes)) {
			different++;
			break;
		}
/*		amt_read += numbytes;
		if (amt_read >= RECORDSIZE) {
			amt_read = 0;
			userec(datarec);
			datarec = findrec();
		}*/
		userec(datarec);
		sparse_ind++;
		size -= numbytes;
	}
	/* 
	 * if the number of bytes read isn't the
	 * number of bytes supposedly in the file,
	 * they're different
	 */
/*	if (amt_read != filesize)
		different++;*/
	userec(datarec);
	free(sparsearray);
	if (different)
		fprintf(msg_file, "%s: data differs\n", filename);

}

/*
 * JK
 * This routine should be used more often than it is ... look into
 * that.  Anyhow, what it does is translate the sparse information
 * on the header, and in any subsequent extended headers, into an
 * array of structures with true numbers, as opposed to character
 * strings.  It simply makes our life much easier, doing so many
 * comparisong and such.
 */
fill_in_sparse_array()
{
	int 	ind;
	long from_oct();

	/*
	 * allocate space for our scratch space; it's initially
	 * 10 elements long, but can change in this routine if
	 * necessary
	 */
	sp_array_size = 10;
	sparsearray = (struct sp_array *) malloc(sp_array_size * sizeof(struct sp_array));

	/*
	 * there are at most five of these structures in the header
	 * itself; read these in first
	 */
	for (ind = 0; ind < SPARSE_IN_HDR; ind++) {
		if (!head->header.sp[ind].numbytes)
			break;
		sparsearray[ind].offset =
			from_oct(1+12, head->header.sp[ind].offset);
		sparsearray[ind].numbytes =
			from_oct(1+12, head->header.sp[ind].numbytes);
	}
	/*
	 * if the header's extended, we gotta read in exhdr's till
	 * we're done
	 */
	if (head->header.isextended) {
 	    /* how far into the sparsearray we are 'so far' */
	    static int so_far_ind = SPARSE_IN_HDR;	
	    union record *exhdr;
   	    
	    for (;;) {
		exhdr = findrec();
		for (ind = 0; ind < SPARSE_EXT_HDR; ind++) {
			if (ind+so_far_ind > sp_array_size-1) {
				/*
 				 * we just ran out of room in our
				 *  scratch area - realloc it
 				 */
				sparsearray = (struct sp_array *)
					realloc(sparsearray, 
						sp_array_size*2*sizeof(struct sp_array));
				sp_array_size *= 2;
			}
			/*
			 * convert the character strings into longs
			 */
			sparsearray[ind+so_far_ind].offset = 
			    from_oct(1+12, exhdr->ext_hdr.sp[ind].offset);
			sparsearray[ind+so_far_ind].numbytes =
			    from_oct(1+12, exhdr->ext_hdr.sp[ind].numbytes);
		}
		/* 
		 * if this is the last extended header for this
		 * file, we can stop
		 */
		if (!exhdr->ext_hdr.isextended)
			break;
		else {
			so_far_ind += SPARSE_EXT_HDR;
			userec(exhdr);
		}
	    }
	    /* be sure to skip past the last one  */
	    userec(exhdr);
	}
}
