/*****************************************************************************
 * $Id: gnu.c,v 2.8 1996/09/23 21:57:50 ak Exp $
 *****************************************************************************
 * $Log: gnu.c,v $
 * Revision 2.8  1996/09/23 21:57:50  ak
 * More buffer space for Posix filenames (NAMSIZ2 instead of NAMSIZ).
 *
 * Revision 2.7  1994/11/08 20:12:22  ak
 * Optional case-sensitive compare.
 * Moved include of port.h to tar.h.
 *
 * Revision 2.6  1994/07/05  14:56:23  edvkai
 * Access Control Lists.
 *
 * Added --substitute, allows redirection of UNC names. No support for links yet.
 *
 * Streamlined absolute path treatment. No skipcrud any more.
 *
 * Bugfix: If a directory existed, EAs were not attached.
 *
 * Revision 2.5  1994/05/31 21:33:27  ak
 * Still there seems to be a memory leak somewhere in EMX _ead_read.
 * Didn't find it with dbmalloc though. Added EA_MODE (former EMX_EA) of
 * 2, using my own ea_* functions to load the attributes and the EMX
 * functions to manipulate them. The leak seems to be gone now. Added
 * --old-ea to allow creating the gtak 2.12 EA records.
 *
 * Revision 2.4  1994/03/20 21:03:26  ak
 * Moved hidden/system and archive bit checks from is_dot_.. to the right
 * place. Reason: GNU dumps should include all files in the directory's
 * file list even if -p is missing, otherwise h/s files could perhaps be
 * deleted on restore.
 *
 * Revision 2.3  1993/12/07 18:39:40  ak
 * Back to own dirent, EMX dirent converts network drives to lowercase.
 *
 * Revision 2.2  1993/09/17  15:10:55  edvkai
 * - Better support for GZip, reorganized compression options.
 * - Buffered mode now available without compression.
 * - Merged GNU dumps and archive bit backups into a single set of options.
 *
 * Revision 2.1  1993/08/08  19:08:17  ak
 * Merge of network TAR with 2.12.
 *
 * Revision 1.11  1993/01/10  11:19:45  ak
 * J.R.: Bugfix: f_gnudump='f' sicherte doppelt.
 *
 * Revision 1.1.1.2  1993/08/08  17:50:22  ak
 * - Network tape access.
 * - Dynamically loaded modules for disk/scsi/network interfaces.
 *
 * Revision 1.10  1992/12/13  09:39:40  ak
 * K.U.R: f_archive/f_reset_archive added to support the "archived" bit.
 *
 * Revision 1.9  1992/10/31  06:55:09  ak
 * Modified -G (again) to specify the generation of increment.
 *
 * Revision 1.8  1992/10/29  10:00:03  ak
 * -G for incremental backups like -g, but w/o writing the dumpfile.
 *
 * Revision 1.7  1992/09/29  09:55:42  ak
 * K.U.R., once again :-)
 * - removed a dup() in buffer.c
 * - EMX opendir with hidden/system test in gnu.c (...dotdot)
 *
 * Revision 1.6  1992/09/26  08:31:50  ak
 * *** empty log message ***
 *
 * Revision 1.5  1992/09/20  17:50:30  ak
 * EMX 0.8 stat() returns random inode numbers. As a result, TAR -g
 * archived every single directory. ifdef'ed out this code.
 *
 * Revision 1.4  1992/09/20  07:47:06  ak
 * Fixes from Kai Uwe Rommel
 *   --checkpoints instead of --semi-verbose (1.11)
 *   -g filenames
 *
 * Revision 1.3  1992/09/12  15:57:30  ak
 * - Usenet patches for GNU TAR 1.10
 * - Bugfixes and patches of Kai Uwe Rommel:
 *         filename conversion for FAT
 *         EMX 0.8e
 *         -0..1 alias for a: b:
 *         -2..7 alias for +++TAPE$x
 *
 * Revision 1.2  1992/09/02  20:08:10  ak
 * Version AK200
 * - Tape access
 * - Quick file access
 * - OS/2 extended attributes
 * - Some OS/2 fixes
 * - Some fixes of Kai Uwe Rommel
 *
 * Revision 1.1.1.1  1992/09/02  19:21:41  ak
 * Original GNU Tar 1.10 with some filenames changed for FAT compatibility.
 *
 * Revision 1.1  1992/09/02  19:21:40  ak
 * Initial revision
 *
 *****************************************************************************/

static char *rcsid = "$Id: gnu.c,v 2.8 1996/09/23 21:57:50 ak Exp $";

/*
 * Modified by Andreas Kaiser July 92.
 * See CHANGES.AK for info.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>

#ifdef BSD42
 #include <sys/dir.h>
#else
 #ifdef __MSDOS__
  #include "msd_dir.h"
 #else
  #ifdef USG
   #ifdef NDIR
    #include <ndir.h>
   #else
    #include <dirent.h>
   #endif
   #ifndef DIRECT
    #define direct dirent
   #endif
   #define DP_NAMELEN(x) strlen((x)->d_name)
  #else
   /*
    * FIXME: On other systems there is no standard place for the header file
    * for the portable directory access routines.  Change the #include line
    * below to bring it in from wherever it is.
    */
   #ifdef OS2
    #include "dirent.h"
    #define direct dirent
    #define DP_NAMELEN(x) strlen((x)->d_name)
    #ifdef OWN_DIRENT
     #define opendir(dir) openxdir(dir,A_HIDDEN|A_SYSTEM)
     /* gnudump directories should include all files, independant of -p */
    #endif
   #else
    #include "ndir.h"
   #endif
  #endif
 #endif
#endif

#ifndef DP_NAMELEN
#define DP_NAMELEN(x)	(x)->d_namlen
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

/*
 * If there are no symbolic links, there is no lstat().  Use stat().
 */
#ifndef S_IFLNK
#define lstat stat
#endif

#ifdef __STDC__
#define VOIDSTAR void *
#else
#define VOIDSTAR char *
#endif

#ifndef S_IFLNK
#define lstat stat
#endif

/* extern VOIDSTAR malloc(); */

#include "tar.h"

extern time_t new_time;
extern FILE *msg_file;

extern VOIDSTAR init_buffer();
extern char *get_buffer();
extern void add_buffer();
extern void flush_buffer();

extern char *new_name();

static void add_dir_name();

struct dirname {
	struct dirname *next;
	char *name;
	char *dir_text;
	int dev;
	int ino;
	int allnew;
};
static struct dirname *dir_list;
static time_t this_time;

void
add_dir(name,dev,ino,text)
char *name;
char *text;
{
	struct dirname *dp;

	dp=(struct dirname *)malloc(sizeof(struct dirname));
	if(!dp)
		abort();
	dp->next=dir_list;
	dir_list=dp;
	dp->dev=dev;
	dp->ino=ino;
	dp->name=malloc(strlen(name)+1);
	strcpy(dp->name,name);
	dp->dir_text=text;
	dp->allnew=0;
}

void
read_dir_file()
{
	int dev;
	int ino;
	char *strp;
	FILE *fp;
	char buf[512];
	extern int errno;
	static char path[MAXPATHLEN];

	if (!gnu_dumpfile)
		return;
	time(&this_time);
#ifdef MSDOS
	if(gnu_dumpfile[0]!='/' && (!isalpha(gnu_dumpfile[0]) || 
				    gnu_dumpfile[1]!=':' || 
				    gnu_dumpfile[2]!='/')) {
		_fullpath(path,gnu_dumpfile,MAXPATHLEN);
		gnu_dumpfile=path;
	}
#else
	if(gnu_dumpfile[0]!='/') {
#ifdef USG
			int getcwd();
			if(!getcwd(path,MAXPATHLEN)) {
				msg("Couldn't get current directory.");
				exit(EX_SYSTEM);
			}
#else
			char *getwd();

			if(!getwd(path)) {
				msg("Couldn't get current directory: %s",path);
				exit(EX_SYSTEM);
			}
#endif
		/* If this doesn't fit, we're in serious trouble */
		strcat(path,"/");
		strcat(path,gnu_dumpfile);
		gnu_dumpfile=path;
	}
#endif

	if (f_gnudump == 'f')
		return;

	fp=fopen(gnu_dumpfile,"r");
	if(fp==0 && errno!=ENOENT) {
		msg_perror("Can't open %s",gnu_dumpfile);
		return;
	}
	if(!fp)
		return;
	fgets(buf,sizeof(buf),fp);
	if(!f_new_files) {
		f_new_files++;
		new_time=atol(buf);
	}
	while(fgets(buf,sizeof(buf),fp)) {
		strp= &buf[strlen(buf)];
		if(strp[-1]=='\n')
			strp[-1]='\0';
		strp=buf;
		dev=atol(strp);
		while(isdigit(*strp))
			strp++;
		ino=atol(strp);
		while(isspace(*strp))
			strp++;
		while(isdigit(*strp))
			strp++;
		strp++;
		add_dir(un_quote_string(strp),dev,ino,(char *)0);
	}
	fclose(fp);
}

void
write_dir_file()
{
	FILE *fp;
	struct dirname *dp;
	char *str;
	extern char *quote_copy_string();

	if (!gnu_dumpfile || f_gnudump == 's')
		return;
	fp=fopen(gnu_dumpfile,"w");
	if(fp==0) {
		msg_perror("Can't write to %s",gnu_dumpfile);
		return;
	}
	fprintf(fp,"%lu\n",this_time);
	for(dp=dir_list;dp;dp=dp->next) {
		if(!dp->dir_text)
			continue;
		str=quote_copy_string(dp->name);
		if(str) {
			fprintf(fp,"%u %u %s\n",dp->dev,dp->ino,str);
			free(str);
		} else
			fprintf(fp,"%u %u %s\n",dp->dev,dp->ino,dp->name);
	}
	fclose(fp);
}

struct dirname *
get_dir(name)
char *name;
{
	struct dirname *dp;

	for(dp=dir_list;dp;dp=dp->next) {
		if(!strcmp(dp->name,name))
			return dp;
	}
	return 0;
}


/* Collect all the names from argv[] (or whatever), then expand them into
   a directory tree, and put all the directories at the beginning. */
collect_and_sort_names()
{
	struct name *n,*n_next;
	int num_names;
	struct stat statbuf;
	int name_cmp();
	char *merge_sort();

	name_gather();

	read_dir_file();
	if(!namelist) addname(".");
	for(n=namelist;n;n=n_next) {
		n_next=n->next;
		if(n->found || n->dir_contents)
			continue;
		if(n->regexp)		/* FIXME just skip regexps for now */
			continue;
		if(n->change_dir)
			if(chdir(n->change_dir)<0) {
				msg_perror("can't chdir to %s",n->change_dir);
				continue;
			}

#ifdef AIX
		if (statx (n->name, &statbuf, STATSIZE, STX_HIDDEN|STX_LINK))
#else
		if(lstat(n->name,&statbuf)<0) {
#endif /* AIX */
			msg_perror("can't stat %s",n->name);
			continue;
		}
		if((statbuf.st_mode&S_IFMT)==S_IFDIR) {
			n->found++;
			add_dir_name(n->name,statbuf.st_dev);
		}
	}

	num_names=0;
	for(n=namelist;n;n=n->next)
		num_names++;
	namelist=(struct name *)merge_sort((VOIDSTAR)namelist,num_names,(char *)(&(namelist->next))-(char *)namelist,name_cmp);

	for(n=namelist;n;n=n->next) {
		n->found=0;
	}
}

int
name_cmp(n1,n2)
struct name *n1,*n2;
{
	if(n1->found) {
		if(n2->found)
			return strcmp(n1->name,n2->name);
		else
			return -1;
	} else if(n2->found)
		return 1;
	else
		return strcmp(n1->name,n2->name);
}

int
dirent_cmp(p1,p2)
char **p1,**p2;
{
	char *frst,*scnd;

	frst= (*p1)+1;
	scnd= (*p2)+1;

	return strcmp(frst,scnd);
}

char *
get_dir_contents(p,device)
char *p;
int device;
{
	DIR *dirp;
	register struct direct *d;
	char *new_buf;
	char *namebuf;
	int bufsiz;
	int len;
	VOIDSTAR the_buffer;
	char *buf;
	int n_strs;
	int n_size;
	char *p_buf;
	char **vec,**p_vec;

	extern int errno;

	errno=0;
	dirp=opendir(p);
	bufsiz=strlen(p)+NAMSIZ2;
	namebuf=ck_malloc(bufsiz+2);
	if(!dirp) {
		if(errno)
			msg_perror("can't open directory %s",p);
		else
			msg("error opening directory %s",p);
		new_buf=NULL;
	} else {
		struct dirname *dp;
		int all_children;

		dp=get_dir(p);
		all_children= dp ? dp->allnew : 0;
		(void) strcpy(namebuf,p);
		if(p[strlen(p)-1]!='/')
			(void) strcat(namebuf,"/");
		len=strlen(namebuf);

		the_buffer=init_buffer();
		while(d=readdir(dirp)) {
			struct stat hs;

			/* Skip . and .. */
			if(is_dot_or_dotdot(d))
				continue;
			if(DP_NAMELEN(d) + len >=bufsiz) {
				bufsiz+=NAMSIZ2;
				namebuf=ck_realloc(namebuf,bufsiz+2);
			}
			(void) strcpy(namebuf+len,d->d_name);
#ifdef AIX
			if (0 != f_follow_links?
			    statx(namebuf, &hs, STATSIZE, STX_HIDDEN):
			    statx(namebuf, &hs, STATSIZE, STX_HIDDEN|STX_LINK))
#else
			if (0 != f_follow_links? stat(namebuf, &hs): lstat(namebuf, &hs))
#endif
			{
				msg_perror("can't stat %s",namebuf);
				continue;
			}
			if(   (f_local_filesys && device!=hs.st_dev)
			   || (f_exclude && check_exclude(namebuf)))
				add_buffer(the_buffer,"N",1);
#ifdef AIX
			else if (S_ISHIDDEN (hs.st_mode)) {
				add_buffer (the_buffer, "D", 1);
				strcat (d->d_name, "A");
				d->d_namlen++;
			}	
#endif /* AIX */
			else if((hs.st_mode&S_IFMT)==S_IFDIR) {
				if(dp=get_dir(namebuf)) {
#ifndef OS2
					if(   dp->dev!=hs.st_dev
 					   || dp->ino!=hs.st_ino) {
					   	if(f_verbose)
							msg("directory %s has been renamed.",namebuf);
						dp->allnew=1;
						dp->dev=hs.st_dev;
						dp->ino=hs.st_ino;
					}
#endif
					dp->dir_text="";
				} else {
					if(f_verbose)
						msg("Directory %s is new",namebuf);
					add_dir(namebuf,hs.st_dev,hs.st_ino,"");
					dp=get_dir(namebuf);
					dp->allnew=1;
				}
				if(all_children)
					dp->allnew=1;

				add_buffer(the_buffer,"D",1);
			} else if(   !all_children
   				&& f_new_files
 				&& new_time>hs.st_mtime
				&& (   f_new_files>1
 				    || new_time>hs.st_ctime))
				add_buffer(the_buffer,"N",1);
			else
				add_buffer(the_buffer,"Y",1);
			add_buffer(the_buffer,d->d_name,(int)(DP_NAMELEN(d)+1));
		}
		add_buffer(the_buffer,"\000\000",2);
		closedir(dirp);

		/* Well, we've read in the contents of the dir, now sort them */
		buf=get_buffer(the_buffer);
		if(buf[0]=='\0') {
			flush_buffer(the_buffer);
			new_buf=NULL;
		} else {
			n_strs=0;
			for(p_buf=buf;*p_buf;) {
				int tmp;

				tmp=strlen(p_buf)+1;
				n_strs++;
				p_buf+=tmp;
			}
			vec=(char **)malloc(sizeof(char *)*(n_strs+1));
			for(p_vec=vec,p_buf=buf;*p_buf;p_buf+=strlen(p_buf)+1)
				*p_vec++= p_buf;
			*p_vec= 0;
			qsort((VOIDSTAR)vec,n_strs,sizeof(char *),dirent_cmp);
			new_buf=(char *)malloc(p_buf-buf+2);
			for(p_vec=vec,p_buf=new_buf;*p_vec;p_vec++) {
				char *p_tmp;

				for(p_tmp= *p_vec;*p_buf++= *p_tmp++;)
					;
			}
			*p_buf++='\0';
			free(vec);
			flush_buffer(the_buffer);
		}
	}
	free(namebuf);
	return new_buf;
}

/* p is a directory.  Add all the files in P to the namelist.  If any of the
   files is a directory, recurse on the subdirectory. . . */
static void
add_dir_name(p,device)
char *p;
int device;
{
	char *new_buf;
	char *p_buf;

	char *namebuf;
	int buflen;
	register int len;
	int sublen;

	VOIDSTAR the_buffer;

	char *buf;
	char **vec,**p_vec;
	int n_strs,n_size;

	struct name *n;

	int dirent_cmp();

	new_buf=get_dir_contents(p,device);

	for(n=namelist;n;n=n->next) {
		if(!strcmp(n->name,p)) {
			n->dir_contents = new_buf;
			break;
		}
	}

	len=strlen(p);
	buflen= NAMSIZ2<=len ? len + NAMSIZ2 : NAMSIZ2;
	namebuf= ck_malloc(buflen+1);

	(void)strcpy(namebuf,p);
	if(namebuf[len-1]!='/') {
		namebuf[len++]='/';
		namebuf[len]='\0';
	}
	for(p_buf=new_buf;p_buf && *p_buf;p_buf+=sublen+1) {
		sublen=strlen(p_buf);
		if(*p_buf=='D') {
			if(len+sublen>=buflen) {
				buflen+=NAMSIZ2;
				namebuf= ck_realloc(namebuf,buflen+1);
			}
			(void)strcpy(namebuf+len,p_buf+1);
			addname(namebuf);
			add_dir_name(namebuf,device);
		}
	}
	free(namebuf);
}

/* Returns non-zero if p is . or ..   This could be a macro for speed. */
is_dot_or_dotdot(d)
struct dirent *d;
{
  	char *p = d->d_name;
	return (p[0]=='.' && (p[1]=='\0' || (p[1]=='.' && p[2]=='\0')));
}






gnu_restore(char *fname)
{
	char *current_dir;
/*	int current_dir_length; */

	char *archive_dir;
/*	int archive_dir_length; */
	VOIDSTAR the_buffer;
	char	*p;
	DIR	*dirp;
	struct direct *d;
	char *cur,*arc;
	extern struct stat hstat;		/* Stat struct corresponding */
	long size,copied;
	char *from,*to;
	extern union record *head;

	dirp=opendir(fname);

	if(!dirp) {
			/* The directory doesn't exist now.  It'll be created.
			   In any case, we don't have to delete any files out
			   of it */
		skip_file((long)hstat.st_size);
		return;
	}

	the_buffer=init_buffer();
	while(d=readdir(dirp)) {
		if(is_dot_or_dotdot(d))
			continue;

		add_buffer(the_buffer,d->d_name,(int)(DP_NAMELEN(d)+1));
	}
	closedir(dirp);
	add_buffer(the_buffer,"",1);

	current_dir=get_buffer(the_buffer);
	archive_dir=(char *)malloc(hstat.st_size);
	if(archive_dir==0) {
		msg("Can't allocate %d bytes for restore",hstat.st_size);
		skip_file((long)hstat.st_size);
		return;
	}
	to=archive_dir;
	for(size=hstat.st_size;size>0;size-=copied) {
		from=findrec()->charptr;
		if(!from) {
			msg("Unexpected EOF in archive\n");
			break;
		}
		copied=endofrecs()->charptr - from;
		if(copied>size)
			copied=size;
		bcopy((VOIDSTAR)from,(VOIDSTAR)to,(int)copied);
		to+=copied;
		userec((union record *)(from+copied-1));
	}

	for(cur=current_dir;*cur;cur+=strlen(cur)+1) {
		for(arc=archive_dir;*arc;arc+=strlen(arc)+1) {
			arc++;
			if(!strcmp(arc,cur))
				break;
		}
		if(*arc=='\0') {
			p=new_name(fname,cur);
			if(f_confirm && !confirm("delete",p)) {
				free(p);
				continue;
			}
			if(f_verbose)
				fprintf(msg_file,"%s: deleting %s\n",tar,p);
			if(recursively_delete(p)) {
				msg("%s: Error while deleting %s\n",tar,p);
			}
			free(p);
		}

	}
	flush_buffer(the_buffer);
	free(archive_dir);
}

recursively_delete(path)
char *path;
{
	struct stat sbuf;
	DIR *dirp;
	struct direct *dp;
	char *path_buf;
	/* int path_len; */


	if(lstat(path,&sbuf)<0)
		return 1;
	if((sbuf.st_mode &S_IFMT)==S_IFDIR) {

		/* path_len=strlen(path); */
		dirp=opendir(path);
		if(dirp==0)
			return 1;
		while(dp=readdir(dirp)) {
			if(is_dot_or_dotdot(dp))
				continue;
			path_buf=new_name(path,dp->d_name);
			if(recursively_delete(path_buf)) {
				free(path_buf);
				closedir(dirp);
				return 1;
			}
			free(path_buf);
		}
		closedir(dirp);

		if(rmdir(path)<0)
			return 1;
		return 0;
	}
	if(unlink(path)<0)
		return 1;
	return 0;
}

